/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.perseus.cache;

import org.objectweb.perseus.cache.api.CacheEntry;
import org.objectweb.perseus.cache.api.CacheException;
import org.objectweb.perseus.cache.api.CacheEventListener;
import org.objectweb.perseus.cache.api.CacheEvent;
import org.objectweb.perseus.cache.lib.BasicCacheManager;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;

public class TestCache extends CacheHelper {

	Object id1, id2, id3, id4;
	Object value1, value2, value3, value4;

	public TestCache(String s) {
		super(s);
		if (logger == null) {
			loggerFactory = Monolog.getMonologFactory();
			logger = loggerFactory
					.getLogger("org.objectweb.perseus.cache.test");
		}
	}

	protected void setUp() throws CacheException {
		id1 = new Integer(1);
		id2 = new Integer(2);
		id3 = new Integer(3);
		id4 = new Integer(4);
		value1 = new Object();
		value2 = new Object();
		value3 = new Object();
		value4 = new Object();
	}

	protected void tearDown() throws Exception {
		cm = null;
		id1 = null;
		id2 = null;
		id3 = null;
		id4 = null;
		value1 = null;
		value2 = null;
		value3 = null;
		value4 = null;
	}

	public void testCache1() throws CacheException {
		setUp(FIFO, 3, 1, 0);
		CacheEntry ce1 = cm.bind(id1, value1);
		CacheEntry ce2 = cm.bind(id2, value2);

		CacheEntry ce = cm.lookup(id1);
		assertNotNull(ce);
		assertTrue(ce.getCeIdentifier() == id1);
		assertTrue(ce.getCeObject() == value1);

		ce = cm.lookup(id3);
		assertNull(ce);

		try {
			cm.bind(id1, value1);
			fail("illegal rebind undetected");
		} catch (CacheException e) {
		}

		try {
			cm.bind(id1, value2);
			fail("illegal rebind undetected");
		} catch (CacheException e) {
		}

		cm.fix(ce1);
		cm.unfix(ce1);

		try {
			cm.unfix(ce2);
			fail("unfix protocol error undetected");
		} catch (CacheException e) {
		}

		cm.touch(ce2);
	}

	public void testFIFOCache() throws CacheException {
		try {
			setUp(FIFO, 3, 1, 0);
			// fills in the cache
			CacheEntry ce1 = cm.bind(id1, value1);
			logger.log(BasicLevel.DEBUG,"ce1=" + ce1);
			CacheEntry ce2 = cm.bind(id2, value2);
			logger.log(BasicLevel.DEBUG,"ce2=" + ce2);
			CacheEntry ce3 = cm.bind(id3, value3);
			logger.log(BasicLevel.DEBUG,"ce3=" + ce3);

			// discard reference to ce1 -> can be garbage collected
			ce1 = null;
			// bind id4 -> ce1 (first in) should be unbound
			CacheEventListener cel = new TestCacheEventListener(id1);
			cm.bindFc(BasicCacheManager.CACHE_LISTENER_BINDING, cel);
			CacheEntry ce4 = cm.bind(id4, value4);
			logger.log(BasicLevel.DEBUG,"ce4=" + ce4);
			cm.unbindFc(BasicCacheManager.CACHE_LISTENER_BINDING);

			// fix ce2
			cm.fix(ce2);
			// touch ce4 -> should not have any effect (FIFO replacement manager)
			cm.touch(ce4);

			// discard reference to ce3 -> can be garbage collected
			ce3 = null;
			// bind id1 -> ce3 (first in) should be unbound (ce2 is older but fixed)
			cel = new TestCacheEventListener(id3);
			cm.bindFc(BasicCacheManager.CACHE_LISTENER_BINDING, cel);
			cm.bind(id1, value1);
			cm.unbindFc(BasicCacheManager.CACHE_LISTENER_BINDING);
		} catch (CacheException e) {
			e.printStackTrace();
			fail("A problem appear with the cache: " + e.getMessage());
		}
	}

	public void testLRUCache() throws CacheException {
		try {
			setUp(LRU, 3, 1, 0);
			CacheEntry ce1 = cm.bind(id1, value1);
			logger.log(BasicLevel.DEBUG,"ce1=" + ce1);
			CacheEntry ce2 = cm.bind(id2, value2);
			logger.log(BasicLevel.DEBUG,"ce2=" + ce2);
			CacheEntry ce3 = cm.bind(id3, value3);
			logger.log(BasicLevel.DEBUG,"ce3=" + ce3);

			// discard reference to ce1 -> can be garbage collected
			ce1 = null;
			// bind id4 -> ce1 (first touched) should be unbound
			CacheEventListener cel = new TestCacheEventListener(id1);
			cm.bindFc(BasicCacheManager.CACHE_LISTENER_BINDING, cel);
			CacheEntry ce4 = cm.bind(id4, value4);
			logger.log(BasicLevel.DEBUG,"ce4=" + ce4);
			cm.unbindFc(BasicCacheManager.CACHE_LISTENER_BINDING);

			// touch cache entries in reverse binding order
			cm.touch(ce4);
			cm.touch(ce3);
			cm.touch(ce2);
			// fix ce4
			cm.fix(ce4);

			// discard reference to ce3 -> can be garbage collected
			ce3 = null;
			// bind id1 -> ce3 (first touched) should be unbound (ce4 is older but fixed)
			cel = new TestCacheEventListener(id3);
			cm.bindFc(BasicCacheManager.CACHE_LISTENER_BINDING, cel);
			cm.bind(id1, value1);
			cm.unbindFc(BasicCacheManager.CACHE_LISTENER_BINDING);
		} catch (CacheException e) {
			e.printStackTrace();
			fail("A problem appear with the cache: " + e.getMessage());
		}
	}

	public void testMRUCache() throws CacheException {
		try {
			setUp(MRU, 3, 1, 0);
			CacheEntry ce1 = cm.bind(id1, value1);
			logger.log(BasicLevel.DEBUG,"ce1=" + ce1);
			CacheEntry ce2 = cm.bind(id2, value2);
			logger.log(BasicLevel.DEBUG,"ce2=" + ce2);
			CacheEntry ce3 = cm.bind(id3, value3);
			logger.log(BasicLevel.DEBUG,"ce3=" + ce3);

			// discard reference to ce3 -> can be garbage collected
			ce3 = null;
			// bind id4 -> ce3 (last touched) should be unbound
			CacheEventListener cel = new TestCacheEventListener(id3);
			cm.bindFc(BasicCacheManager.CACHE_LISTENER_BINDING, cel);
			CacheEntry ce4 = cm.bind(id4, value4);
			logger.log(BasicLevel.DEBUG,"ce4=" + ce4);
			cm.unbindFc(BasicCacheManager.CACHE_LISTENER_BINDING);

			// touch cache entries in reverse binding order
			cm.touch(ce4);
			cm.touch(ce2);
			cm.touch(ce1);
			// fix ce1
			cm.fix(ce1);

			// discard reference to ce2 -> can be garbage collected
			ce2 = null;
			// bind id3 -> ce2 (last touched) should be unbound (ce1 is newer but fixed)
			cel = new TestCacheEventListener(id2);
			cm.bindFc(BasicCacheManager.CACHE_LISTENER_BINDING, cel);
			cm.bind(id3, value3);
			cm.unbindFc(BasicCacheManager.CACHE_LISTENER_BINDING);
		} catch (CacheException e) {
			e.printStackTrace();
			fail("A problem appear with the cache: " + e.getMessage());
		}
	}

	public void testCacheFull() throws CacheException {
		setUp(FIFO, 3, 1, 0);
		cm.fix(cm.bind(id1, value1));
		cm.fix(cm.bind(id2, value2));
		cm.fix(cm.bind(id3, value3));
		try {
			cm.bind(id4, value4);
//			fail("Cache full error undetected");
		} catch (CacheException e) {
		}
	}

	static class TestCacheEventListener implements CacheEventListener {
		private Object expected;

		public TestCacheEventListener(Object expected) {
			this.expected = expected;
		}

		public void entryBound(CacheEvent event) {
		}

		public void entryUnbound(CacheEvent event) {
			if (event.getCeIdentifier() != expected) {
				throw new RuntimeException(
						"Unexpected entry unbound " + event.getCeIdentifier());
			}
		}
	}

	public void testCacheThreshold() throws CacheException {
		setUp(FIFO, 3, 1, 1);
		CacheEntry ce1 = cm.bind(id1, value1);
		CacheEntry ce2 = cm.bind(id2, value2);
		ce1 = null;
		Thread.yield();
		try {
			Thread.sleep(500);
		} catch (InterruptedException e) {
		}
		Object o = cm.lookup(id1);
		assertNull("element not unbound: " + o, o);
		cm.getBgcleaner().stop();
		try {
			cm.getBgcleaner().getThread().join(1000);
		} catch (InterruptedException e) {
		}
		assertTrue("The Background cleaner is aliver",
				!cm.getBgcleaner().getThread().isAlive());
	}
	
	public void testCacheThreshold2() throws CacheException {
		setUp(FIFO, 3, 1, 2);
		cm.bind(id1, value1);
		cm.bind(id2, value2);
		Thread.yield();
		Object o = cm.lookup(id1);
		assertNotNull("element not unbound: " + o, o);
		cm.getBgcleaner().stop();
		try {
			cm.getBgcleaner().getThread().join(1000);
		} catch (InterruptedException e) {
		}
		assertTrue("The Background cleaner is aliver",
				!cm.getBgcleaner().getThread().isAlive());
	}
}
